# -*- coding: utf-8 -*-
"""
Created on Wed Jul 25 11:23:14 2018

Pyene Devices collects the data, location and connections of the different
devices considered in pyene

@author: Dr Eduardo Alejandro Martínez Ceseña
https://www.researchgate.net/profile/Eduardo_Alejandro_Martinez_Cesena
"""
import math
import numpy as np
import copy
from pyomo.core import Constraint

'''                          CONFIGURATION CLASSES                          '''


class BusConfig:
    ''' Default settings for an electricity bus '''
    def __init__(self):
        # Basic settings
        aux = ['BASE_KV', 'BS', 'BUS_AREA', 'BUS_TYPE', 'BUS_X', 'BUS_Y',
               'Demand', 'GS', 'PeakP', 'PeakQ', 'Position', 'Name', 'Number',
               'VM', 'VA', 'VMAX', 'VMIN', 'ZONE', 'Load_Type']
        self.settings = {}
        for x in aux:
            self.settings[x] = None

    def MPCconfigure(self, mpc, No=0):
        ''' Configure using mat power data '''

        self.settings['Position'] = No
        self.settings['Number'] = mpc['BUS_I'][No]
        self.settings['PeakP'] = mpc['PD'][No]
        self.settings['PeakP'] = mpc['QD'][No]

        aux = ['BASE_KV', 'BS', 'BUS_AREA', 'BUS_TYPE', 'GS', 'VM', 'VA',
               'VMAX', 'VMIN', 'ZONE']
        for x in aux:
            self.settings[x] = mpc[x][No]

        #  Optional data - not included in all files
        if 'BUS_NAME' in mpc.keys():
            if mpc['BUS_NAME'] != []:
                self.settings['Name'] = mpc['BUS_NAME'][No]
        if 'BUS_X' in mpc.keys():
            self.settings['BUS_X'] = mpc['BUS_X'][No]
            self.settings['BUS_Y'] = mpc['BUS_Y'][No]
        if 'Load_Type' in mpc.keys():
            self.settings['Load_Type'] = mpc['Load_Type'][No]
        if 'Loss_Fix' in mpc.keys():
            self.settings['Loss_Fix'] = mpc['Loss_Fix'][No]
        else:
            self.settings['Loss_Fix'] = 0


class BranchConfig:
    ''' Electricity Branch '''
    def __init__(self):
        # Basic settings
        aux = ['ANGMAX', 'ANGMIN', 'BR_B', 'BR_R', 'BR_STATUS', 'BR_X',
               'Number', 'F_BUS', 'RATE_A', 'RATE_A', 'RATE_C', 'TAP',
               'T_BUS', 'Loss_Fix', 'MTTF', 'MTTR']
        self.settings = {}
        for x in aux:
            self.settings[x] = None

    def MPCconfigure(self, mpc, No=0):
        ''' Configure using mat power data '''

        self.settings['Number'] = No+1
        self.settings['Position'] = No

        aux = ['ANGMAX', 'ANGMIN', 'BR_B', 'BR_R', 'BR_STATUS', 'BR_X',
               'F_BUS', 'RATE_A', 'RATE_B', 'RATE_C', 'TAP', 'T_BUS', 'MTTF',
               'MTTR']
        for x in aux:
            if x in mpc:
                self.settings[x] = mpc[x][No]

        if 'Loss_Fix' in mpc.keys():
            self.settings['Loss_Fix'] = mpc['Loss_Fix'][No]
        else:
            self.settings['Loss_Fix'] = 0


class ConventionalConfig:
    ''' Conventnional generator '''
    def __init__(self):
        # Basic settings
        aux = ['Ancillary', 'APF', 'GEN_BUS', 'MBASE', 'PC1', 'PC2',
               'PG', 'PMAX', 'PMIN', 'QC1MIN', 'QC1MAX', 'QC2MIN', 'QC2MAX',
               'QG', 'QMAX', 'QMIN', 'Ramp', 'RAMP_AGC', 'RAMP_10', 'RAMP_30',
               'RAMP_Q', 'RES', 'VG', 'MDT', 'MUT', 'GEN', 'MTTF', 'MTTR']
        self.settings = {}
        for x in aux:
            self.settings[x] = None

        # Cost data
        aux = ['COST', 'MODEL', 'NCOST', 'SHUTDOWN', 'STARTUP']
        self.cost = {}
        for x in aux:
            self.cost[x] = None

    def MPCconfigure(self, mpc, conv, No=0):
        ''' Configure using mat power data '''

        # Generator settings - from mat power file
        self.settings['Position'] = No
        aux = ['APF', 'GEN_BUS', 'MBASE', 'PC1', 'PC2', 'PG', 'PMAX',
               'PMIN', 'QC1MIN', 'QC1MAX', 'QC2MIN', 'QC2MAX', 'QG', 'QMAX',
               'QMIN', 'RAMP_AGC', 'RAMP_10', 'RAMP_30', 'RAMP_Q', 'VG',
               'GEN', 'MTTF', 'MTTR']
        for x in aux:
            if x in mpc['gen']:
                self.settings[x] = mpc['gen'][x][No]

        # Generator costs - from mat power file
        aux = ['COST', 'MODEL', 'NCOST', 'SHUTDOWN', 'STARTUP']
        for x in aux:
            self.cost[x] = mpc['gencost'][x][No]

        # Ramp constraints
        aux = len(conv['Ramp'])
        if aux == 0:  # No values from settings
            if 'Ramp' in mpc['gen'].keys():
                # Values can be taken from settings
                self.settings['Ramp'] = mpc['gen']['Ramp'][No]
            else:
                self.settings['Ramp'] = None
        elif aux == 1:  # Assign same ramp to all generators
            self.settings['Ramp'] = conv['Ramp'][0]
        else:  # Assign a ramp to each generator
            if No <= aux:
                self.settings['Ramp'] = conv['Ramp'][No]
            else:  # But the information was not provided
                self.settings['Ramp'] = None

        # MUT and MDT constraints
        aux = ['MUT', 'MDT']
        for x in aux:
            auxL = len(conv[x])
            if auxL == 0:
                self.settings[x] = None
            elif auxL == 1:
                self.settings[x] = conv[x][0]
            else:
                self.settings[x] = conv[x][No]

        # Provision of services
        aux = ['Ancillary', 'Baseload', 'RES']
        for x in aux:
            if len(conv[x]) == 1:
                self.settings[x] = conv[x][0]
            else:
                self.settings[x] = conv[x][No]


class HydropowerConfig:
    ''' Hydropower generator '''
    def __init__(self):
        # Basic settings
        aux = ['Ancillary', 'Baseload', 'Bus', 'Max', 'Ramp',
               'RES', 'Position', 'Min']
        self.settings = {}
        for x in aux:
            self.settings[x] = None

        aux = ['MODEL', 'NCOST', 'COST']
        self.cost = {}
        for x in aux:
            self.cost[x] = None

    def MPCconfigure(self, hydro, No=0):
        ''' Configure using hydropower settings '''
        self.settings['Position'] = No
        aux = ['Ancillary', 'Baseload', 'Ramp', 'RES']
        for x in aux:
            if len(hydro[x]) == 1:
                self.settings[x] = hydro[x][0]
            else:
                self.settings[x] = hydro[x][No]

        self.settings['Bus'] = hydro['Bus'][No]
        self.settings['Max'] = hydro['Max'][No]
        if hydro['Min'] != []:
            self.settings['Min'] = hydro['Min'][No]
        else:
            self.settings['Min'] = 0
        
        aux = ['MUT', 'MDT']
        for x in aux:
            self.settings[x] = None

        # Default cost model
        self.cost['MODEL'] = 1
        self.cost['NCOST'] = 2
        self.cost['COST'] = [0, 0, 1, hydro['Cost'][No]]


class RESConfig:
    ''' RES generator '''
    def __init__(self):
        # Basic settings
        aux = ['Bus', 'Cost', 'Max', 'Uncertainty', 'Position']
        self.settings = {}
        for x in aux:
            self.settings[x] = None

        aux = ['MODEL', 'NCOST', 'COST']
        self.cost = {}
        for x in aux:
            self.cost[x] = None

    def MPCconfigure(self, RES, No=0):
        ''' Configure using RES settings '''
        self.settings['Position'] = No
        self.settings['Bus'] = RES['Bus'][No]
        self.settings['Max'] = RES['Max'][No]
        self.settings['Uncertainty'] = RES['Uncertainty']

        # Default cost model
        self.cost['MODEL'] = 1
        self.cost['NCOST'] = 2
        self.cost['COST'] = [0, 0, 1, RES['Cost'][No]]


'''                               DEVICE CLASSES                            '''


class Branch:
    ''' Electricity branch '''
    def __init__(self, obj):
        ''' Initialise bus class

        The class can use the following parameters:
        ['ANGMAX', 'ANGMIN', 'BR_STATUS', 'Number',
        'F_BUS', 'Position', 'RATE_A', 'RATE_B', 'RATE_C', 'T_BUS']
        However, only the ones that are currently used are passed
        '''

        aux = ['BR_R', 'BR_X', 'F_BUS', 'Position', 'RATE_A', 'T_BUS', 'TAP',
               'BR_B', 'Loss_Fix', 'BR_STATUS', 'MTTF', 'MTTR']

        # Get settings
        self.data = {}
        for xa in aux:
            self.data[xa] = obj.settings[xa]

        self.data['F_Position'] = None
        self.data['T_Position'] = None

        # Paremeters used for pyomo
        self.pyomo = {}
        self.pyomo['N-1'] = None

    def cNDCLossA_rule(self, m, xt, xL, ConF, ConL, A, B, Bse):
        ''' Power losses estimation - Positive '''
        return m.vNLoss[ConL+self.get_Pos(), xt] >= \
            self.data['Loss_Fix']/Bse + \
            (A[xL]+B[xL]*m.vNFlow[ConF+self.get_Pos(), xt]) * \
            self.data['BR_R']

    def cNDCLossB_rule(self, m, xt, xL, ConF, ConL, A, B, Bse):
        ''' Power losses estimation - Negative '''
        return m.vNLoss[ConL+self.get_Pos(), xt] >= \
            self.data['Loss_Fix']/Bse + \
            (A[xL]-B[xL]*m.vNFlow[ConF+self.get_Pos(), xt]) * \
            self.data['BR_R']

    def cNEFlow_rule(self, m, xt, xs, ConF, ConV, Bus):
        ''' Set DC power flow constraint '''
        if self.is_active(xs):
            xaux1 = ConV+Bus[self.get_PosF()].get_Sec(xs)
            xaux2 = ConV+Bus[self.get_PosT()].get_Sec(xs)

            return m.vNFlow[ConF+self.get_Sec(xs), xt] == \
                (m.vNVolt[xaux1, xt]-m.vNVolt[xaux2, xt])/self.data['BR_X']
        else:
            return Constraint.Skip

    def cNEFMax_rule(self, m, xt, xs, ConF):
        ''' Branch capacity constraint (positive) '''
        if self.is_active(xs):
            return m.vNFlow[ConF+self.get_Sec(xs), xt] >= \
                -self.data['RATE_A']
        else:
            return Constraint.Skip

    def cNEFMin_rule(self, m, xt, xs, ConF):
        ''' Branch capacity constraint (positive) '''
        if self.is_active(xs):
            return m.vNFlow[ConF+self.get_Sec(xs), xt] <= \
                self.data['RATE_A']
        else:
            return Constraint.Skip

    def get_B(self):
        ''' Get Susceptance '''
        return self.data['BR_B']

    def get_BusF(self):
        ''' Get bus number at beginning (from) of the branch '''
        return self.data['F_BUS']

    def get_BusT(self):
        ''' Get bus number at end (to) of the branch '''
        return self.data['T_BUS']

    def getLoss(self):
        ''' Return non technical losses in the bus '''
        return self.data['Loss_Fix']

    def get_N1(self, x=':'):
        ''' Get values for a single N-1 condition '''
        return self.pyomo['N-1'][x]

    def get_Number(self):
        ''' Get branch number - starting from one'''
        return self.data['Position']+1

    def get_Pos(self):
        ''' Get position of the branch - starting from zero'''
        return self.data['Position']

    def get_PosF(self):
        ''' Get bus position at beginning (from) of the branch '''
        return self.data['F_Position']

    def get_PosT(self):
        ''' Get bus position at end (to) of the branch '''
        return self.data['T_Position']

    def get_R(self):
        ''' Get Resistance '''
        return self.data['BR_R']

    def get_Rate(self):
        ''' Get Rate A for normal operation conditions'''
        return self.data['RATE_A']

    def get_Sec(self, xs=':'):
        ''' Get position in N-1 scenario '''
        return self.pyomo['N-1'][xs]

    def get_Tap(self):
        ''' Get tap position '''
        return self.data['TAP']

    def get_X(self):
        ''' Get Reactance '''
        return self.data['BR_X']

    def is_active(self, xs):
        ''' Is the line connected in this scenario? '''
        return self.pyomo['N-1'][xs] is not None

    def set_B(self, val):
        ''' Set Susceptance'''
        self.data['BR_B'] = val

    def set_PosF(self, val):
        ''' Set bus position at beginning (from) of the branch '''
        self.data['F_Position'] = val

    def set_PosT(self, val):
        ''' Set bus position at end (to) of the branch '''
        self.data['T_Position'] = val

    def set_N1(self, val, x=None):
        ''' Set values for all conditions '''
        if x is None:
            self.pyomo['N-1'] = val
        else:
            self.pyomo['N-1'][x] = val

    def set_R(self, val):
        ''' Set Resistance'''
        self.data['BR_R'] = val

    def set_Rate(self, val):
        ''' Set Rate A for normal operation conditions'''
        self.data['RATE_A'] = val

    def set_X(self, val):
        ''' Set Reactance'''
        self.data['BR_X'] = val


class Bus:
    ''' Electricity bus '''
    def __init__(self, obj):
        ''' Initialise bus class

        The class can use the following parameters:
        [, 'BS', 'BUS_AREA', 'BUS_X','BUS_Y','Demand',
        'GS', 'PeakP', 'PeakQ', 'Position', 'Name', 'Number', 'VA',
        , 'ZONE']
        However, only the ones that are currently used are passed
        '''
        # Parameters currently in use
        aux = ['BUS_X', 'BUS_Y', 'Demand', 'PeakP', 'PeakQ', 'Position',
               'Name', 'Number', 'BUS_TYPE', 'BASE_KV', 'VMAX', 'VMIN', 'VM',
               'Load_Type', 'Loss_Fix']

        # Get settings
        self.data = {}
        for xa in aux:
            self.data[xa] = obj.settings[xa]

        # New data
        if self.data['Load_Type'] is None:
            self.data['Load_Type'] = 0  # 0) Urban 1) Rural
        self.data['F_Branches'] = []  # Branches connected from the bus
        self.data['T_Branches'] = []  # Branches connected to the bus
        self.data['F_Loss'] = []  # Branches connected from the bus - Losses
        self.data['T_Loss'] = []  # Branches connected to the bus - Losses
        self.data['NoFB'] = 0  # Number of branches connected from the bus
        self.data['NoTB'] = 0  # Number of branches connected to the bus
        self.data['GenType'] = []  # Types of generators connected to the bus
        self.data['GenPosition'] = []  # Position of the generators

        self.pyomo = {}
        self.pyomo['N-1'] = None

    def add_BraF(self, val):
        ''' Append value to F_Branches - Branches connected from node'''
        self.data['F_Branches'].append(val)
        self.data['NoFB'] += 1

    def add_BraT(self, val):
        ''' Append value to T_Branches - Branches connected to node'''
        self.data['T_Branches'].append(val)
        self.data['NoTB'] += 1

    def add_Gen(self, xt, xp):
        ''' Append generator type and position '''
        self.data['GenType'].append(xt)
        self.data['GenPosition'].append(xp)

    def get_BraF(self):
        ''' Get list of branches connected from the bus in an N-1 scenario'''
        return self.data['F_Branches']

    def get_BraT(self):
        ''' Get list of branches connected to the bus in an N-1 scenario'''
        return self.data['T_Branches']

    def get_kV(self):
        ''' Get base kV '''
        return self.data['BASE_KV']

    def get_LossF(self):
        ''' Get list of branches connected from the bus - Losses'''
        return self.data['F_Loss']

    def get_LossT(self):
        ''' Get list of branches connected to the bus - Losses'''
        return self.data['T_Loss']

    def getLoss(self):
        ''' Return non technical losses in the bus '''
        return self.data['Loss_Fix']

    def get_LT(self):
        ''' Get load type (1:Urban, 2:Rural) '''
        return self.data['Load_Type']

    def get_Number(self):
        ''' Get Bus number '''
        return self.data['Number']

    def get_Pos(self):
        ''' Get Bus position - beginning from zero '''
        return self.data['Position']

    def get_Sec(self, xs=':'):
        ''' Get position of variable in N-1 scenario '''
        return self.pyomo['N-1'][xs]

    def get_Type(self):
        ''' Get bus type '''
        return self.data['BUS_TYPE']

    def get_VM(self):
        ''' Get Voltege magnitude (pu) '''
        return self.data['VM']

    def get_Vmax(self):
        ''' Get max voltage limit (pu) '''
        return self.data['VMAX']

    def get_Vmin(self):
        ''' Get max voltage limit (pu) '''
        return self.data['VMIN']

    def get_X(self):
        ''' Get X coordinates '''
        return self.data['BUS_X']

    def get_Y(self):
        ''' Get Y coordinates '''
        return self.data['BUS_Y']

    def set_LossF(self, val):
        ''' Set list of branches connected from the bus - Losses'''
        self.data['F_Loss'] = val

    def set_LossT(self, val):
        ''' Set list of branches connected to the bus - Losses'''
        self.data['T_Loss'] = val

    def set_N1(self, val, x=None):
        ''' Set values for all conditions '''
        if x is None:
            self.pyomo['N-1'] = val
        else:
            self.pyomo['N-1'][x] = val

    def set_LT(self, val):
        ''' Set load type (0:Urban, 1:Rural) '''
        self.data['Load_Type'] = val


class ElectricityNetwork:
    ''' Electricity network '''
    def __init__(self, NoBus=1, NoBranch=1):
        ''' General electricity network settings '''
        self.data = {
                'version': None,
                'baseMVA': None,
                'NoGen': None,
                'Slack': None,
                'Buses': NoBus,  # Number of buses
                'Branches': NoBranch,  # Number of branches
                'Security': None,  # N-1 cases to consider
                'SecurityNo': None  # Number of N-1 cases
                }
        self.loss = {}
        self.loss['A'] = None
        self.loss['B'] = None

        # Define bus objects - configuration class
        self.BusConfig = [BusConfig() for x in range(NoBus)]

        # Define branch objects - configuration class
        self.BranchConfig = [BranchConfig() for x in range(NoBranch)]

    def cNDCLossA_rule(self, m, xt, xb, xL, ConF, ConL):
        ''' Power losses estimation - Positive '''
        return self.Branch[xb].cNDCLossA_rule(m, xt, xL, ConF, ConL,
                                              self.loss['A'], self.loss['B'],
                                              self.get_Base())

    def cNDCLossB_rule(self, m, xt, xb, xL, ConF, ConL):
        ''' Power losses estimation - Negative '''
        return self.Branch[xb].cNDCLossB_rule(m, xt, xL, ConF, ConL,
                                              self.loss['A'], self.loss['B'],
                                              self.get_Base())

    def cNEFlow_rule(self, m, xt, xb, xs, ConF, ConV):
        ''' Branch flows constraint '''
        return self.Branch[xb].cNEFlow_rule(m, xt, xs, ConF, ConV, self.Bus)

    def cNEFMax_rule(self, m, xt, xb, xs, ConF):
        ''' Branch capacity constraint (positive) '''
        return self.Branch[xb].cNEFMax_rule(m, xt, xs, ConF)

    def cNEFMin_rule(self, m, xt, xb, xs, ConF):
        ''' Branch capacity constraint (positive) '''
        return self.Branch[xb].cNEFMin_rule(m, xt, xs, ConF)

    def findBusPosition(self, bus):
        ''' Find the position of a bus

        This is required as the bus numbers may not begin from zero, or some
        positions may be missing
        '''
        xn = 0
        while self.Bus[xn].data['Number'] != bus:
            xn += 1

        return xn

    def get_Base(self):
        ''' Provide base MVA rating '''
        return self.data['baseMVA']

    def get_FlowF(self, xn, xs):
        ''' Get branches connected from bus per scenario '''
        aux = []
        for xb in self.Bus[xn].get_BraF():  # Branches connected to the bus
            # Is teh branch active in the scenario?
            if self.Branch[xb].get_N1(xs) is not None:
                aux.append(self.Branch[xb].get_N1(xs))
        return aux
        aux = []

    def get_FlowT(self, xn, xs):
        ''' Get branches connected to bus per scenario '''
        aux = []
        for xb in self.Bus[xn].get_BraT():  # Branches connected to the bus
            # Is teh branch active in the scenario?
            if self.Branch[xb].get_N1(xs) is not None:
                aux.append(self.Branch[xb].get_N1(xs))
        return aux

    def get_NoBra(self):
        ''' Get total number of branches in the network '''
        return self.data['Branches']

    def get_NoBus(self):
        ''' Get total number of buses in the network '''
        return self.data['Buses']

    def initialise(self, sett):
        ''' Prepare objects and remove configuration versions '''

        # Initialise bus object
        self.Bus = [Bus(self.BusConfig[x]) for x in
                    range(self.data['Buses'])]
        del self.BusConfig

        # Initialise branch object
        self.Branch = [Branch(self.BranchConfig[x]) for x in
                       range(self.data['Branches'])]
        del self.BranchConfig

        # Security constraints
        if sett['SecurityFlag']:  # Consider all N-1 constraints
            self.data['SecurityNo'] = self.data['Branches']
            self.data['N-1'] = range(self.data['Branches'])
        else:
            self.data['SecurityNo'] = len(sett['Security'])
            self.data['N-1'] = sett['Security']

        # Match buses and nodes
        xcou = 0
        for ob in self.Branch:
            # Find position of the bus (may be different from the number)
            xf = self.findBusPosition(ob.get_BusF())
            xt = self.findBusPosition(ob.get_BusT())

            # The branch now includes the position of the buses
            ob.set_PosF(xf)
            ob.set_PosT(xt)

            # Enable branch for the first N-1 scenario (intact network)
            ob.set_N1([None]*(self.data['SecurityNo']+1))
            ob.set_N1(xcou, 0)
            xcou += 1

            # Tbe bus now includes the position of the relevant branches
            self.Bus[xf].add_BraF(ob.get_Pos())
            self.Bus[xt].add_BraT(ob.get_Pos())

            # Adjust line capacity
            ob.set_Rate(ob.get_Rate()/self.data['baseMVA'])

        # Initialize security data for nodes
        for ob in self.Bus:
            ob.set_N1([None]*(self.data['SecurityNo']+1))
            ob.set_N1(ob.get_Pos(), 0)

        # Are all the loads the same type?
        aux = len(sett['Load_Type'])
        if aux == 1:
            if sett['Load_Type'][0] == 1:
                # An update is only needed if the loads are rural
                for ob in self.Bus:
                    ob.set_LT(sett['Load_Type'])
        elif aux > 1:
            # Update a set of the buses
            xb = 0
            for val in sett['Load_Type']:
                self.Bus[xb].set_LT(val)
                xb += 1

        # Enable branches in other scenarios (pyomo)
        xsec = 0
        for xs in self.data['N-1']:
            xsec += 1
            # Add N-1 information to buses
            for ob in self.Bus:
                ob.set_N1(xsec*self.data['Buses']+ob.data['Position'], xsec)

            # Add N-1 information to branches
            for ob in (self.Branch[xb] for xb in range(self.data['Branches'])
                       if xb+1 != xs):
                ob.set_N1(xcou, xsec)
                xcou += 1

        # Model losses
        if sett['Losses']:
            for ob in self.Bus:
                ob.set_LossF(ob.get_BraF())
                ob.set_LossT(ob.get_BraT())

    def MPCconfigure(self, mpc):
        ''' Initialize using mat power data '''

        # General electricity network settings
        for xa in ['version', 'baseMVA', 'NoGen', 'Slack']:
            self.data[xa] = mpc[xa]

        # Bus data
        for x in range(mpc['NoBus']):
            self.BusConfig[x].MPCconfigure(mpc['bus'], x)

        # Branch data
        for x in range(mpc['NoBranch']):
            self.BranchConfig[x].MPCconfigure(mpc['branch'], x)


class GenClass:
    ''' Core generation class '''
    def cNEBaseload_rule(self, m, xt, Tim, w, ConG):
        ''' Impose provision of baseload '''
        if self.data['Baseload'] > 0:
            return m.vNGen[ConG+self.pyomo['vNGen'], xt] >= \
                sum(m.vNGen[ConG+self.pyomo['vNGen'], xt2] for xt2 in Tim) * \
                self.data['Baseload']*w
        else:
            return Constraint.Skip

    def cNEGenC_rule(self, m, xc, xt, ConC, ConG, w):
        ''' Piece wise cost estimation '''
        (flg, x1,  x2, M1, M2) = self.cNEGenC_Auxrule(xc, ConC, ConG)
        if flg:
            return m.vNGCost[x1, xt]/w >= m.vNGen[x2, xt]*M1 + M2
        return Constraint.Skip

    def cNEGenC_Auxrule(self, xc, ConC, ConG):
        ''' Auxiliary for cNEGenC_rule '''
        flg = xc < self.pyomo['NoPieces']
        if flg:
            x1 = ConC+self.pyomo['vNGen']
            x2 = ConG+self.pyomo['vNGen']
            M1 = self.cost['LCost'][xc][0]
            M2 = self.cost['LCost'][xc][1]
        else:
            x1 = 0
            x2 = 0
            M1 = 0
            M2 = 0

        return flg, x1, x2, M1, M2

    def cNEGMax_rule(self, m, xt, ConG):
        ''' Maximum generation capacity '''
        return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= self.data['Max']

    def cNEGMaxUC_rule(self, m, xt, ConG):
        ''' Maximum generation capacity '''
        if self.pyomo['vNGen_Bin'] is None:
            return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= self.data['Max']

        return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= self.data['Max'] * \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_Bin'], xt]

    def cNEGMaxUCWC_rule(self, m, xt, ConG, Der):
        ''' Maximum generation capacity '''
        ''' Maximum generation capacity '''
        if self.pyomo['vNGen_Bin'] is None:
            return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= self.data['Max']

        return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= self.data['Max'] * \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_Bin'], xt]*Der[xt]

    def cNEGMaxWC_rule(self, m, xt, ConG, Der):
        ''' Maximum generation capacity '''
        return m.vNGen[ConG+self.pyomo['vNGen'], xt] <= \
            self.data['Max']*Der[xt]

    def cNEGMin_rule(self, m, xt, ConG):
        ''' Minimum generation capacity - Only when needed '''
        if self.data['Max'] > 0:
            return m.vNGen[ConG+self.pyomo['vNGen'], xt] >= self.data['Min']
        else:
            return Constraint.Skip

    def cNEGMinDT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum down time '''
        if self.pyomo['vNGen_MDT'] is None:
            return Constraint.Skip

        return m.vNGen_MDT[ConG+self.pyomo['vNGen_MDT'], xt] >= \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_MDT'], xt0] - \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_MDT'], xt]

    def cNEGMinDT2_rule(self, m, xt, LL, ConG):
        ''' Minimum down time '''
        if self.pyomo['vNGen_MDT'] is None:
            return Constraint.Skip

        return m.vNGen_MDT[ConG+self.pyomo['vNGen_MDT'], xt]*self.data['MDT'] \
            <= sum(1-m.vNGen_Bin[ConG+self.pyomo['vNGen_MDT'], xL] for xL in
                   LL[xt-1:xt+self.data['MDT']-1])

    def cNEGMinUC_rule(self, m, xt, ConG):
        ''' Minimum generation capacity - Only when needed '''
        if self.pyomo['vNGen_Bin'] is None:
            return Constraint.Skip

        return m.vNGen[ConG+self.pyomo['vNGen'], xt] >= \
            self.data['Min']*m.vNGen_Bin[ConG+self.pyomo['vNGen_Bin'], xt]

    def cNEGMinUCWC_rule(self, m, xt, ConG, Der):
        ''' Minimum generation capacity '''
        if self.pyomo['vNGen_Bin'] is None:
            return Constraint.Skip

        return m.vNGen[ConG+self.pyomo['vNGen'], xt] >= \
            self.data['Min']*m.vNGen_Bin[ConG+self.pyomo['vNGen_Bin'], xt] * \
            Der[xt]

    def cNEGMinUT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum up time '''
        if self.pyomo['vNGen_MUT'] is None:
            return Constraint.Skip

        return m.vNGen_MUT[ConG+self.pyomo['vNGen_MUT'], xt] >= \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_MUT'], xt] - \
            m.vNGen_Bin[ConG+self.pyomo['vNGen_MUT'], xt0]

    def cNEGMinUT2_rule(self, m, xt, LL, ConG):
        ''' Minimum up time - Only when needed '''
        if self.pyomo['vNGen_MUT'] is None:
            return Constraint.Skip

        return m.vNGen_MUT[ConG+self.pyomo['vNGen_MUT'], xt]*self.data['MUT'] \
            <= sum(m.vNGen_Bin[ConG+self.pyomo['vNGen_MUT'], xL] for xL in
                   LL[xt-1:xt+self.data['MUT']-1])

    def cNGenRampDown_rule(self, m, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        if self.data['Ramp'] < self.data['Max']:
            aux = ConG+self.pyomo['vNGen']
            return m.vNGen[aux, xt]-m.vNGen[aux, xt0] <= self.data['Ramp']
        else:
            return Constraint.Skip

    def cNGenRampUp_rule(self, m, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        if self.data['Ramp'] < self.data['Max']:
            aux = ConG+self.pyomo['vNGen']
            return m.vNGen[aux, xt0]-m.vNGen[aux, xt] <= self.data['Ramp']
        else:
            return Constraint.Skip

    def get_Bin(self):
        ''' get binaries for UC '''
        return self.pyomo['vNGen_Bin']

    def get_Bus(self):
        ''' Get bus number '''
        return self.data['Bus']

    def get_BusPos(self):
        ''' Get bus position '''
        return self.data['BusPosition']

    def get_Max(self):
        ''' Get maximum capacity (MW) '''
        return self.data['Max']

    def get_Min(self):
        ''' Get minimum capacity (MW) '''
        return self.data['Min']

    def get_NoPieces(self):
        ''' Get number of pices used for piece-wise cost estimations '''
        return self.pyomo['NoPieces']

    def get_P(self):
        ''' Get power output '''
        return self.data['PG']

    def get_Q(self):
        ''' Get reactive power output '''
        return self.data['QG']

    def get_Pos(self):
        ''' Get generator position '''
        return self.data['Position']

    def get_VG(self):
        ''' Get voltage magnitude'''
        return self.data['VG']

    def get_vNGen(self):
        ''' Set position of vNGen variable - pyomo'''
        return self.pyomo['vNGen']

    def initialise(self):
        ''' Finalize initialization '''
        # Adjust ramps
        if self.data['Ramp'] is None:
            self.data['Ramp'] = self.data['Max']+1
        elif self.data['Ramp'] <= 0 or self.data['Ramp'] >= 1:
            self.data['Ramp'] = self.data['Max']+1
        else:
            self.data['Ramp'] = self.data['Max']*self.data['Ramp']

    def set_Bin(self, xbin):
        ''' Set binaries for UC '''
        if self.data['Max'] > 0 and self.data['Min'] > 0:
            self.pyomo['vNGen_Bin'] = xbin
            xbin += 1
        else:
            self.pyomo['vNGen_Bin'] = None

        return xbin

    def set_MDT(self, xMDT):
        ''' Set variables for minimum down time '''
        if self.pyomo['vNGen_Bin'] is None or self.data['MDT'] is None:
            return xMDT

        self.pyomo['vNGen_MDT'] = xMDT
        xMDT += 1

        return xMDT

    def set_MUT(self, xMUT):
        ''' Set variables for minimum up time '''
        if self.pyomo['vNGen_Bin'] is None or self.data['MUT'] is None:
            return xMUT

        self.pyomo['vNGen_MUT'] = xMUT
        xMUT += 1

        return xMUT

    def set_CostCurve(self, sett, xNo, xLen, Base):
        ''' Define piece wise cost curve approximation '''

        if self.cost['MODEL'] == 1:
            # Piece-wise model
            NoPieces = self.cost['NCOST']
            xval = np.zeros(NoPieces, dtype=float)
            yval = np.zeros(NoPieces, dtype=float)
            xp = 0
            for x in range(NoPieces):
                xval[x] = self.cost['COST'][xp]
                yval[x] = self.cost['COST'][xp+1]
                xp += 2
            NoPieces -= 1
        else:
            # Polinomial model

            # Select number of pieces for the approximation
            if xLen == 0:  # Default case
                Delta = self.data['Max']
                if(self.cost['COST'][0] != 0):
                    Delta /= 3
            elif xLen == 1:  # Single value for all generators
                Delta = sett['Pieces'][0]
            else:  # Predefined number of pieces
                Delta = sett['Pieces'][xNo]

            NoPieces = int(np.floor(self.data['Max']/Delta))
            xval = np.zeros(NoPieces+1, dtype=float)
            yval = np.zeros(NoPieces+1, dtype=float)
            aux = self.data['Min']
            for xp in range(NoPieces+1):
                xval[xp] = aux
                xc = self.cost['NCOST']-1
                yval[xp] = self.cost['COST'][xc]
                for x in range(1, self.cost['NCOST']):
                    xc -= 1
                    yval[xp] += self.cost['COST'][xc]*xval[xp]**x
                aux += Delta

        # Convert to LP constraints
        self.cost['LCost'] = np.zeros((NoPieces, 2), dtype=float)
        for xv in range(NoPieces):
            self.cost['LCost'][xv][0] = (yval[xv+1]-yval[xv]) / (xval[xv+1] -
                                                                 xval[xv])
            self.cost['LCost'][xv][1] = \
                yval[xv]-xval[xv]*self.cost['LCost'][xv][0]
            self.cost['LCost'][xv][0] *= Base

        self.pyomo['NoPieces'] = NoPieces

    def set_Max(self, val):
        ''' Set maximum capacity (MW) '''
        self.data['Max'] = val

    def set_Min(self, val):
        ''' Set minimum capacity (MW) '''
        self.data['Min'] = val

    def set_PosB(self, xb):
        ''' Set position of the bus '''
        self.data['BusPosition'] = xb

    def set_vNGen(self, val):
        ''' Set position of vNGen variable - pyomo'''
        self.pyomo['vNGen'] = val


class Conventional(GenClass):
    ''' Conventional generator '''
    def __init__(self, obj):
        ''' Initialise generator class

        The class can use the following parameters:
        ['APF', 'MBASE', 'PC1', 'PC2', 'PG', 'QC1MIN', 'QC1MAX',
        'QC2MIN', 'QC2MAX', 'QG', 'RAMP_AGC',
        'RAMP_10', 'RAMP_30', 'RAMP_Q', 'RES', 'VG']
        However, only the ones that are currently used are passed
        ['COST', 'MODEL', 'NCOST', 'SHUTDOWN', 'STARTUP']
        '''
        # Parameters currently in use
        aux = ['Ancillary', 'Baseload', 'PMAX', 'PMIN', 'Ramp', 'Position',
               'VG', 'PG', 'QG', 'MDT', 'MUT', 'QMAX', 'QMIN', 'GEN', 'MTTF',
               'MTTR']

        # Get settings
        self.data = {}
        self.data['Max'] = obj.settings['PMAX']
        self.data['Min'] = obj.settings['PMIN']
        for xa in aux:
            self.data[xa] = obj.settings[xa]
        self.data['Bus'] = obj.settings['GEN_BUS']


        aux = ['COST', 'MODEL', 'NCOST', 'SHUTDOWN', 'STARTUP']
        self.cost = {}
        for xa in aux:
            self.cost[xa] = obj.cost[xa]

        self.pyomo = {}
        self.pyomo['vNGen'] = None
        self.pyomo['vNGen_Bin'] = None
        self.pyomo['vNGen_MDT'] = None
        self.pyomo['vNGen_MUT'] = None
        self.pyomo['NoPieces'] = None


class Hydropower(GenClass):
    ''' Hydropower generator '''
    def __init__(self, obj):
        ''' Initialise hydropower generator class

        The class can use the following parameters:
        ['Ancillary', 'Baseload', 'Bus', 'Max', 'Ramp', 'RES', 'Position']
        ['COST', 'MODEL', 'NCOST', 'SHUTDOWN', 'STARTUP']
        However, only the ones that are currently used are passed
        '''
        # Parameters currently in use
        aux = ['Ancillary', 'Baseload', 'Bus', 'Max', 'Ramp', 'RES',
               'Position', 'Min', 'MDT', 'MUT']

        # Get settings
        self.data = {}
        for xa in aux:
            self.data[xa] = obj.settings[xa]

        aux = ['COST', 'MODEL', 'NCOST']
        self.cost = {}
        for xa in aux:
            self.cost[xa] = obj.cost[xa]

        self.pyomo = {}
        self.pyomo['vNGen'] = None
        self.pyomo['vNGen_Bin'] = None
        self.pyomo['vNGen_MDT'] = None
        self.pyomo['vNGen_MUT'] = None
        self.pyomo['NoPieces'] = None

    def get_Cost(self):
        ''' Return linear costs '''
        return self.cost['COST'][3]

    def cNEGMinDT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum down time '''
        return Constraint.Skip

    def cNEGMinDT2_rule(self, m, xt, LL, ConG):
        ''' Minimum down time '''
        return Constraint.Skip

    def cNEGMinUT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum up time '''
        return Constraint.Skip

    def cNEGMinUT2_rule(self, m, xt, LL, ConG):
        ''' Minimum up time '''
        return Constraint.Skip


class RES(GenClass):
    ''' RES generation '''
    def __init__(self, obj):
        ''' Initialise RES generator class

        The class can use the following parameters:
        ['Bus', 'Cost', 'Max', 'Uncertainty', 'Position']
        ['MODEL', 'NCOST', 'COST']
        However, only the ones that are currently used are passed
        '''
        # Parameters currently in use
        aux = ['Bus', 'Cost', 'Max', 'Uncertainty', 'Position']

        # Get settings
        self.data = {}
        self.data['Min'] = 0
        for xa in aux:
            self.data[xa] = obj.settings[xa]

        aux = ['MODEL', 'NCOST', 'COST']
        self.cost = {}
        for xa in aux:
            self.cost[xa] = obj.cost[xa]

        self.pyomo = {}
        self.pyomo['vNGen'] = None
        self.pyomo['NoPieces'] = None

    def cNEBaseload_rule(self, m, xt, Tim, w, ConG):
        ''' Impose provision of baseload '''
        return Constraint.Skip

    def cNEGMax_rule(self, m, xt, ConG):
        ''' Maximum generation capacity '''
        return Constraint.Skip

    def cNEGMaxUC_rule(self, m, xt, ConG):
        ''' Maximum generation capacity '''
        return Constraint.Skip

    def cNEGMin_rule(self, m, xt, ConG):
        ''' Minimum generation capacity '''
        return Constraint.Skip

    def cNEGMinDT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum down time '''
        return Constraint.Skip

    def cNEGMinDT2_rule(self, m, xt, LL, ConG):
        ''' Minimum down time '''
        return Constraint.Skip

    def cNEGMinUC_rule(self, m, xt, ConG):
        ''' Minimum generation capacity '''
        return Constraint.Skip

    def cNEGMinUT1_rule(self, m, xt, xt0, ConG):
        ''' Minimum up time '''
        return Constraint.Skip

    def cNEGMinUT2_rule(self, m, xt, LL, ConG):
        ''' Minimum up time '''
        return Constraint.Skip

    def cNGenRampDown_rule(self, m, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        return Constraint.Skip

    def cNGenRampUp_rule(self, m, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        return Constraint.Skip

    def get_Bin(self):
        ''' get binaries for UC '''
        return None

    def get_Cost(self):
        ''' Return linear costs '''
        return self.cost['COST'][3]

    def set_Bin(self, xbin):
        ''' Set binaries for UC '''
        return xbin

    def set_MDT(self, xMDT):
        ''' Set variables for minimum down time '''
        return xMDT

    def set_MUT(self, xMUT):
        ''' Set variables for minimum up time '''
        return xMUT

    def initialise(self):
        ''' Finalize initialization '''
        return


class Generators:
    ''' Electricity generators '''
    def __init__(self, NoConv=0, NoHydro=0, NoRES=0):
        ''' General generator settings '''
        self.data = {
                'Conv': NoConv,
                'Hydro': NoHydro,
                'RES': NoRES,
                'Gen': NoConv+NoHydro+NoRES,
                'Types': None,
                'Bin': None
                }
        self.cooling = {}
        self.cooling['Flag'] = False
        self.cooling['Gen2Der'] = []
        self.cooling['Derate'] = None

        self.pyomo = {}
        self.pyomo['NoPieces'] = 0  # Max number of piece-wise cost curves
        self.pyomo['Type'] = []
        self.pyomo['Pos'] = []

        # Conventional generators
        self.ConvConf = [ConventionalConfig() for x in range(NoConv)]

        # Hydropower generators
        self.HydroConf = [HydropowerConfig() for x in range(NoHydro)]

        # RES generators
        self.RESConf = [RESConfig() for x in range(NoRES)]

    def cNEBaseload_rule(self, m, xg, xt, Tim, w, wF, ConG):
        ''' Baseload rule '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEBaseload_rule(m, xt, Tim, w[xt]/wF,
                                                      ConG)

    def _GClass(self, xg):
        ''' Get class and position of generator corresponsing to xg '''
        xa = self.data['Types'][self.pyomo['Type'][xg]]
        xp = self.pyomo['Pos'][xg]

        return (xa, xp)

    def cNEGenC_rule(self, m, xg, xc, xt, ConC, ConG, w):
        ''' Generation costs - Piece-wise estimation '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGenC_rule(m, xc, xt, ConC, ConG, w)

    def cNEGenC_Auxrule(self, xg, xc, ConC, ConG):
        ''' Auxiliar for Generation costs - Piece-wise estimation '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGenC_Auxrule(xc, ConC, ConG)

    def cNEGMax_rule(self, m, xg, xt, ConG):
        ''' Maximum generation capacity '''
        (xa, xp) = self._GClass(xg)
        # Considering water cooling constraints for conventional generators
        if self.cooling['Flag'] and self.pyomo['Type'][xg] == 0:
            aux = self.cooling['Derate'][self.cooling['Gen2Der'][xp]]
            return getattr(self, xa)[xp].cNEGMaxWC_rule(m, xt, ConG, aux)
        else:  # Ignoring water cooling constraints
            return getattr(self, xa)[xp].cNEGMax_rule(m, xt, ConG)

    def cNEGMaxUC_rule(self, m, xg, xt, ConG):
        ''' Maximum generation capacity '''
        (xa, xp) = self._GClass(xg)
        # Considering water cooling constraints for conventional generators
        if self.cooling['Flag'] and self.pyomo['Type'][xg] == 0:
            aux = self.cooling['Derate'][self.cooling['Gen2Der'][xp]]
            return getattr(self, xa)[xp].cNEGMaxUCWC_rule(m, xt, ConG, aux)
        else:  # Ignoring water cooling constraints
            return getattr(self, xa)[xp].cNEGMaxUC_rule(m, xt, ConG)

    def cNEGMin_rule(self, m, xg, xt, ConG):
        ''' Minimum generation capacity '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMin_rule(m, xt, ConG)

    def cNEGMinDT1_rule(self, m, xg, xt, xt0, ConG):
        ''' Minimum down time '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMinDT1_rule(m, xt, xt0, ConG)

    def cNEGMinDT2_rule(self, m, xg, xt, LL, ConG):
        ''' Minimum down time '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMinDT2_rule(m, xt, LL, ConG)

    def cNEGMinUC_rule(self, m, xg, xt, ConG):
        ''' Minimum generation capacity '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMinUC_rule(m, xt, ConG)

    def cNEGMinUT1_rule(self, m, xg, xt, xt0, ConG):
        ''' Minimum up time '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMinUT1_rule(m, xt, xt0, ConG)

    def cNEGMinUT2_rule(self, m, xg, xt, LL, ConG):
        ''' Minimum up time '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNEGMinUT2_rule(m, xt, LL, ConG)

    def cNGenRampDown_rule(self, m, xg, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNGenRampDown_rule(m, xt, xt0, ConG)

    def cNGenRampUp_rule(self, m, xg, xt, xt0, ConG):
        ''' Generation ramps (down)'''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].cNGenRampUp_rule(m, xt, xt0, ConG)

    def get_Bin(self, xg):
        ''' Get position of binaries for a generator '''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].get_Bin()

    def get_GenInBus(self, Bus):
        ''' Get list of generators connected to a bus - vNGen'''
        aux = []
        for xt, xp in zip(Bus.data['GenType'], Bus.data['GenPosition']):
            aux.append(getattr(self, self.data['Types'][xt])[xp].get_vNGen())

        return aux

    def get_GenAll(self):
        ''' Get all nodes - vNGen '''
        aux = self.get_GenDataAll('pyomo', 'vNGen')

        return aux

    def get_GenAllC(self):
        ''' Get RES nodes - vNGen '''
        aux = self.get_GenDataType('pyomo', 'vNGen', 'Conv')

        return aux

    def get_GenAllH(self):
        ''' Get RES nodes - vNGen '''
        aux = self.get_GenDataType('pyomo', 'vNGen', 'Hydro')

        return aux

    def get_GenAllR(self):
        ''' Get RES nodes - vNGen '''
        aux = self.get_GenDataType('pyomo', 'vNGen', 'RES')

        return aux

    def get_GenDataAll(self, at='data', dt='Bus'):
        ''' Get piece of data from all generators '''
        Dat = np.zeros(self.data['Gen'], dtype=int)
        x = 0
        for ax in self.data['Types']:
            for ob in getattr(self, ax):
                Dat[x] = getattr(ob, at)[dt]
                x += 1

        return Dat

    def get_GenDataType(self, at='data', dt='Bus', tp='Hydro'):
        ''' Get piece of data from specific generators '''
        Dat = np.zeros(self.data[tp], dtype=int)
        x = 0
        for ob in getattr(self, tp):
            Dat[x] = getattr(ob, at)[dt]
            x += 1

        return Dat

    def get_Max(self, xg):
        ''' Generation capacity'''
        (xa, xp) = self._GClass(xg)
        return getattr(self, xa)[xp].get_Max()

    def get_NoBin(self):
        ''' Number of binaries required for the generators '''
        return self.data['Bin']

    def get_NoCon(self):
        ''' Number of generators '''
        return self.data['Conv']

    def get_NoGen(self):
        ''' Number of generators '''
        return self.data['Gen']

    def get_NoHydro(self):
        ''' Number of hydropower units '''
        return self.data['Hydro']

    def get_NoMDT(self):
        ''' Number of binaries required for the generators '''
        return self.data['MDT']

    def get_NoMUT(self):
        ''' Number of binaries required for the generators '''
        return self.data['MUT']

    def get_NoPieces(self):
        ''' Return number of pieces for cost estimations  '''
        return self.pyomo['NoPieces']

    def get_NoRES(self):
        ''' Number of RES units '''
        return self.data['RES']

    def get_vNGenH(self, xg):
        ''' Get position of vNGen variable - pyomo/hydro'''
        return getattr(self, 'Hydro')[xg].get_vNGen()

    def get_vNGenR(self, xg):
        ''' Get position of vNGen variable - pyomo/hydro'''
        return getattr(self, 'RES')[xg].get_vNGen()

    def initialise(self, ENetwork, sett, RM):
        ''' Prepare objects and remove configuration versions '''

        # Get cooling information
        self.cooling['Flag'] = RM.cooling['Flag']
        if self.cooling['Flag']:
            # Link generators and capacity derate profiles
            self.cooling['Gen2Der'] = np.zeros(self.data['Conv'], dtype=int)
            for x in range(len(RM.cooling['Gen2Temp'])):
                self.cooling['Gen2Der'][x] = RM.cooling['Gen2Temp'][x]

            # NUmber of profiles and time periods
            NoT = len(RM.cooling['temp_w'][0])

            # Link generator types to cooling data
            aux = len(RM.cooling['GenType'])
            if aux == 0:
                # All generator types are the same
                NoW = len(RM.cooling['temp_w'])
                # Produce profiles
                self.cooling['Derate'] = np.zeros((NoW, NoT), dtype=float)
                for x in range(NoW):
                    self.cooling['Derate'][x] = \
                        RM.get_CF(temp_w=RM.cooling['temp_w'][x], index=0)
            else:
                # Check generator types
                axType = np.zeros(self.data['Conv'], dtype=int)
                for x in range(aux):
                    axType[x] = RM.cooling['GenType'][x]
                # Find number of simulations to be made,
                # i.e., unique combinations of temperatures and generator types
                aux = {}
                aux['No'] = 0
                aux['type'] = []
                aux['prof'] = []
                for x1 in range(self.data['Conv']):
                    x2 = 0
                    # Check if it is a new or repeated case
                    while x2 <= x1:
                        if x2 == x1:  # Is it the end of the loop?
                            # New profile
                            aux['type'].append(axType[x1])
                            aux['prof'].append(self.cooling['Gen2Der'][x1])
                            self.cooling['Gen2Der'][x1] = aux['No']
                            aux['No'] += 1
                        elif axType[x1] == axType[x2] and \
                                self.cooling['Gen2Der'][x1] == \
                                self.cooling['Gen2Der'][x2]:
                            # Same as x2
                            self.cooling['Gen2Der'][x1] = x2
                            x2 = x1
                        x2 += 1
                # Produce profiles
                self.cooling['Derate'] = \
                    np.zeros((aux['No'], NoT), dtype=float)
                for x in range(aux['No']):
                    self.cooling['Derate'][x] = \
                        RM.get_CF(temp_w=RM.cooling['temp_w'][aux['prof'][x]],
                                  index=aux['type'][x])

        # Initialise conventional generation object
        self.Conv = [Conventional(self.ConvConf[x]) for x in
                     range(self.data['Conv'])]
        del self.ConvConf

        # Initialise hydropower generator objects
        self.Hydro = [Hydropower(self.HydroConf[x]) for x in
                      range(self.data['Hydro'])]
        del self.HydroConf

        # Initialize RES generator objects
        self.RES = [RES(self.RESConf[x]) for x in range(self.data['RES'])]
        del self.RESConf

        # Initialise generators
        self.data['Types'] = ['Conv', 'Hydro', 'RES']
        xLen = len(sett['Pieces'])
        xbin = 0
        xMDT = 0
        xMUT = 0
        xt = 0
        xNo = 0
        for ax in self.data['Types']:
            xp = 0
            for ob in getattr(self, ax):
                # Link generators and buses
                xb = ENetwork.findBusPosition(ob.get_Bus())  # Bus
                xp = ob.data['Position']  # Generator
                # The Generator knows the position of its bus
                ob.set_PosB(xb)
                # The bus knows the type and location of the generator
                ENetwork.Bus[xb].add_Gen(xt, xp)

                # Add UC considerations
                xbin = ob.set_Bin(xbin)
                xMDT = ob.set_MDT(xMDT)
                xMUT = ob.set_MUT(xMUT)

                # Create cost curves
                ob.set_CostCurve(sett, xNo, xLen, ENetwork.get_Base())

                # MW --> pu
                ob.set_Max(ob.get_Max()/ENetwork.get_Base())
                ob.set_Min(ob.get_Min()/ENetwork.get_Base())

                # Store location of vGen variable
                ob.set_vNGen(xNo)

                # Link between positions an generation classes
                self.pyomo['Type'].append(xt)
                self.pyomo['Pos'].append(xp)

                # Store maximum number of pieces used so far
                aux = ob.get_NoPieces()
                if aux > self.pyomo['NoPieces']:
                    self.pyomo['NoPieces'] = aux

                # Finalise initialisation
                ob.initialise()

                xp += 1
                xNo += 1
            xt += 1
        self.data['Bin'] = xbin
        self.data['MUT'] = xMUT
        self.data['MDT'] = xMDT

    def MPCconfigure(self, mpc, conv, hydro, RES):
        ''' Initialize using mat power data '''

        # Conventional generators
        for x in range(self.data['Conv']):
            self.ConvConf[x].MPCconfigure(mpc, conv, x)

        for x in range(self.data['Hydro']):
            self.HydroConf[x].MPCconfigure(hydro, x)

        for x in range(self.data['RES']):
            self.RESConf[x].MPCconfigure(RES, x)

    def set_Max(self, xg, val):
        ''' Update generation capacity '''
        (xa, xp) = self._GClass(xg)
        getattr(self, xa)[xp].set_Max(val)



import os
from pyene.engines.pyene import pyeneConfig


def json_directory():
    ''' Directory contain JSON files for pytest '''
    return os.path.join(os.path.dirname(__file__), '..', 'tests', 'json')


def fixed_config():
    '''Dedicated configuration for pytests '''
    conf = pyeneConfig()
    #TODO: Don't load any file by default
    conf.EM.settings['File'] = os.path.join(json_directory(),
                                            'TreePreload.json')
    conf.NM.settings['File'] = os.path.join(json_directory(), 'case4.json')

    # Hydropower
    conf.NM.hydropower['Number'] = 0  # Number of hydropower plants
    conf.NM.hydropower['Bus'] = []  # Location (bus) of hydro
    conf.NM.hydropower['Max'] = []  # Generation capacity
    conf.NM.hydropower['Cost'] = []  # Costs

    # Pumps
    conf.NM.pumps['Number'] = 0  # Number of pumps
    conf.NM.pumps['Bus'] = []  # Location (bus) of pumps
    conf.NM.pumps['Max'] = []  # Generation capacity
    conf.NM.pumps['Value'] = []  # Value/Profit

    # RES generators
    conf.NM.RES['Number'] = 0  # Number of RES generators
    conf.NM.RES['Bus'] = []  # Location (bus) of pumps
    conf.NM.RES['Max'] = []  # Generation capacity
    conf.NM.RES['Cost'] = []  # Costs

    # Network considerations
    conf.NM.settings['Security'] = []  # List of contingescies to test
    conf.NM.settings['Losses'] = False  # Model losses
    conf.NM.settings['Feasibility'] = False  # Add dummy generators
    conf.NM.scenarios['Weights'] = None  # Weights for each time step
    conf.NM.settings['NoTime'] = 1  # Number of time steps

    # Scenarios
    conf.NM.scenarios['NoDem'] = 2  # Number of demand profiles
    conf.NM.scenarios['NoRES'] = 2  # Number of RES profiles

    # Hydrology
    conf.HM.settings['Flag'] = False  # Disable pyeneH

    return conf


def testConfig():
    return fixed_config()



import os
import sys
import platform
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext as _build_ext


def setup_package():
    compiler_directives = {
        "language_level": "3str",
        "embedsignature": True
    }

    annotate = False

    class new_build_ext(_build_ext):
        def finalize_options(self):
            # Defer the import of Cython and NumPy until after setup_requires
            from Cython.Build.Dependencies import cythonize
            import numpy

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=compiler_directives,
                annotate=annotate,
            )
            if not self.include_dirs:
                self.include_dirs = []
            elif isinstance(self.include_dirs, str):
                self.include_dirs = [self.include_dirs]
            self.include_dirs.append(numpy.get_include())
            super().finalize_options()

    metadata = dict(
        name="pyensys",
        version='0.0.3',
        url='https://github.com/jnmelchorg/pyensys',
        description='python energy systems simulator - pyensys.',
        author='Dr. Eduardo Alejandro Martínez Ceseña, \
            Dr. Jose Nicolas Melchor Gutierrez',
        author_email='Eduardo.MartinezCesena@manchester.ac.uk, \
            jose.melchorgutierrez@manchester.ac.uk',
        packages=find_packages(),
        package_data={'pyensys': ['json/*.json']},
        install_requires=['click', 'pandas', 'pyomo', 'pypsa'],
        extras_require={"test": ["pytest"]},
        cmdclass={"build_ext": new_build_ext},
        # use_scm_version=True,
        entry_points='''
        [console_scripts]
        pyensys=pyensys.cli:cli
        ''',
        classifiers=[
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: C++",
            "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
            "Operating System :: Microsoft :: Windows :: Windows 10"
        ],
    )
    
    config = parse_optional_arguments()

    annotate = config["annotate"]

    if config["profile"]:
        compiler_directives["profile"] = True

    if config["trace"]:
        compiler_directives["linetrace"] = True
    
    metadata["ext_modules"] = ext_modules = []

    if config["glpk"]:
        ext_modules.append(Extension("pyensys.engines.cython._glpk", ["pyensys/engines/_glpk.pyx"],
        include_dirs=[findglpkheaderpath()] if findglpkheaderpath() is not None else [],
        library_dirs=[findglpklibrarypath()] if findglpklibrarypath() is not None else [],
        libraries=["glpk"],))

    if config["clp"]:
        if platform.system() == "Windows":
            ext_modules.append(Extension("pyensys.engines.cython.cpp_energy_wrapper", ["pyensys/engines/cpp_energy_wrapper.pyx"],
            include_dirs=[
                      os.path.dirname(os.path.abspath(__file__))+'\pyensys\engines\external_files\\boost_1_75_0',
                      os.path.dirname(os.path.abspath(__file__))+"\pyensys\engines\external_files\Clp",],
            libraries=['libClp', 'libCoinUtils'],
            library_dirs=[os.path.dirname(os.path.abspath(__file__))+"\pyensys\engines\external_files\Clp\lib"],
                            ))
        elif platform.system() == "Linux":
            ext_modules.append(Extension("pyensys.engines.cython.cpp_energy_wrapper", ["pyensys/engines/cpp_energy_wrapper.pyx"],
            libraries=['Clp'],
            ))
    setup(**metadata)

def parse_optional_arguments():
    config = {
        "glpk": True,
        "clp": True,
        "annotate": False,
        "profile": False,
        "trace": False,
    }

    if "--without-glpk" in sys.argv:
        config["glpk"] = False
        sys.argv.remove("--without-glpk")
    
    if "--without-clp" in sys.argv:
        config["clp"] = False
        sys.argv.remove("--without-clp")

    if "--annotate" in sys.argv:
        config["annotate"] = True
        sys.argv.remove("--annotate")

    if "--enable-profiling" in sys.argv:
        config["profile"] = True
        sys.argv.remove("--enable-profiling")

    if "--enable-trace" in sys.argv:
        config["trace"] = True
        sys.argv.remove("--enable-trace")
    return config

def findglpkheaderpath():
    inc_arg = "-I"
    for arg in sys.argv:
        if arg.startswith(inc_arg) and len(arg) > len(inc_arg):
            return arg[len(inc_arg):]

    # Finding glpk header path
    pythonpath = os.path.split(sys.executable)[0]
    if len(pythonpath.rsplit('/b', 1)) > 1:
        aux1 = pythonpath.rsplit('/b', 1)[1]
        aux2 = pythonpath.rsplit('/b', 1)[0]
        print(aux1)
        print(aux2)
        if aux1 == 'in':
            pythonpath = aux2
    trypaths = [pythonpath+'/Library/include/glpk.h',
                pythonpath+'/include/glpk.h',
                pythonpath+'/Library/include/glpk.h',
                pythonpath+'/include/glpk.h']
    glpkpath = None
    for paths in trypaths:
        if os.path.isfile(paths):
            glpkpath = paths[:-7]
            break
    if glpkpath is None:
        print('Path for GLPK header has not been found in the predefined \
            directories')
        return
    
    return glpkpath

def findglpklibrarypath():   
    lib_arg = "-L"
    for arg in sys.argv:
        if arg.startswith(lib_arg) and len(arg) > len(lib_arg):
            return arg[len(lib_arg):]

    # Finding glpk header path
    pythonpath = os.path.split(sys.executable)[0]
    if len(pythonpath.rsplit('/b', 1)) > 1:
        aux1 = pythonpath.rsplit('/b', 1)[1]
        aux2 = pythonpath.rsplit('/b', 1)[0]
        print(aux1)
        print(aux2)
        if aux1 == 'in':
            pythonpath = aux2
    trypaths = [pythonpath+'/Library/lib/glpk.lib',
                pythonpath+'/libs/glpk.lib',
                pythonpath+'/Library/lib/libglpk.so',
                pythonpath+'/lib/libglpk.so']
    glpkpath = None
    for paths in trypaths:
        if os.path.isfile(paths):
            glpkpath = paths[:-9]
            break
    if glpkpath is None:
        print('Path for GLPK library has not been found in the predefined \
            directories')
        return
    
    return glpkpath

if __name__ == "__main__":
    setup_package()



# Function method to interface the energy model in c++ with python

# Created by Dr. Jose Nicolas Melchor Gutierrez

from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool

#cdef extern from "energy_models.h":
#    cdef cppclass models_cpp:
#        models_cpp() except +
#        void add_bus(vector[double] &P, int n, string t)
#        void add_branch(vector[double] &r, vector[double] &x, 
#            vector[double] &Pmax, int fr, int to, int n, string t)
#        void add_generator(vector[double] &Pmax, 
#            vector[double] &Pmin, int bn, int n, string t,
#            double fc, double vc, vector[double] &a_pwl, 
#            vector[double] &b_pwl, bool is_active)
#        void set_integer_data_power_system(string name, int value)
#        void set_continuous_data_power_system(string name, double value)
#        void run_reduced_dc_opf()
#        void get_generation_solution(vector[double] &generation,
#            vector[double] &generation_cost)
#        void get_branch_solution(vector[double] &power_flow)
#        void get_node_solution(vector[double] &angle, 
#            vector[double] &generation_curtailment, 
#            vector[double] &load_curtailment)
#        double get_objective_function_nm()
#
#        void load_energy_tree_information(int number_nodes, int number_tree, 
#            vector[vector[int] ] &LLEB,
#            vector[vector[int] ] &LLEA,
#            vector[double] &energy_intake,
#            vector[double] &energy_output,
#            vector[double] &weight_nodes)
#        void run_energy_tree()
#        void get_energy_tree_solution(vector[double] &PartialStorage, 
#            vector[double] &TotalStorage, vector[double] &InputsTree,
#            vector[double] &OutputsTree)
#        double get_objective_function_em()
#
#        void load_combined_energy_dc_opf_information(
#            vector[ vector[int] ] &LLNA, vector[int] &CTG)
#        void run_combined_energy_dc_opf_r1()
#        double get_objective_function_combined_energy_dc_opf_r1()
#
#        void run_iterative_reduced_dc_opf()
#
#        void run_iterative_reduced_dc_opf_v2()
#
#        void run_iterative_combined_energy_dc_opf()
#
#        void run_iterative_combined_energy_dc_opf_v2()
#        
#
#cdef extern from "energy_models.cpp":
#    pass

cdef extern from "energy_models2.h":
    cdef cppclass models:
        models_cpp() except +
        void create_parameter()
        void load_double(const string& na, const double& val, const bool& is_vector);
        void load_integer(const string& na, const int& val, const bool& is_vector);
        void load_bool(const string& na, const bool& val, const bool& is_vector);
        void load_string(const string& na, const string& val, const bool& is_vector);
        void set_parameter(const string& typ);
        int update_parameter();
        void get_MOEA_variables(vector[string]& IDs, vector[string]& names, vector[double]& min, vector[double]& max);
        void get_moea_objectives(vector[string]& names);

        void initialise();
        void evaluate();

        void return_outputs(vector[double]& values, vector[int]& starts, vector[ vector[ vector[ string] ] ]& characteristics);

cdef extern from "energy_models2.cpp":
    pass


"""
A common template of parameters for power systems and temporal trees is provided
in this file. Most of the initial functions have been taken from the file pyeneD
created by Dr Eduardo Alejandro Martínez Ceseña.

@author: Dr. Jose Nicolas Melchor Gutierrez
         Dr Eduardo Alejandro Martínez Ceseña
"""

import copy

'''                               DEVICE CLASSES                            '''

class _CommonMethods():
    def __init__(self):
        ''' This Class contains the methods that are common to some classes '''
        # Assign random value to self._data in this function. The value will 
        # be overwritten in the child classes
        self._data = {}

    def get_element(self, name=None, pos=':'):
        ''' This function returns the value or a list of values for the 
        requested parameter "name". 
        - If the parameter is a list then the function returns either a list 
        with the positions indicated in "pos" or the value of the position "pos"
        - "pos" needs to indicate a position between zero and the size of the 
        list
        - If the parameters is a value then "pos" can be ignored'''
        auxp = "No valid name has been passed to the function get_element in \
            the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self._data:
            if isinstance(self._data[name], list):
                if pos == ':':
                    return self._data[name]
                elif isinstance(pos, list):
                    aux = []
                    for aux1 in pos:
                        aux.append(self._data[name][aux1])
                    return aux
                else:
                    return self._data[name][pos]
            else:
                return self._data[name]
        else:
            auxp = "No valid key {0} has been passed to the function \
            get_element in the class {1}. The valid keys for this class are as \
            follows: \n {2}".format(name, self.__class__.__name__, \
            self._data.keys()) 
            assert " ".join(auxp.split())
    
    def get_no_elements(self, name=None):
        ''' This function returns the number of elements in a list '''
        auxp = "No valid name has been passed to the function get_no_elements \
            in the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self._data:
            if isinstance(self._data[name], list):
                return len(self._data[name])
            else:
                auxp = "the key {0} in the class {1} is not a list".format(\
                    name, self.__class__.__name__)
                assert name is not None, " ".join(auxp.split())
        else:
            auxp = "No valid key {0} has been passed to the function \
            get_no_elements in the class {1}. The valid keys for this class are \
            as follows: \n {2}".format(name, self.__class__.__name__, \
            self._data.keys()) 
            assert " ".join(auxp.split())
    
    def set_element(self, name=None, val=None):
        ''' This function set the value or a list of values ("val") to the 
        requested parameter "name". 
        - This function rewrites the existing values or value stored in the 
        parameter "name" '''
        auxp = "No valid name has been passed to the function get_element in \
            the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self._data:
            if isinstance(self._data[name], list):
                auxp = "You are trying to set the \
                parameter {0} that have to be a list with different type of \
                data element in the class {1}".format(name, \
                self.__class__.__name__)
                assert isinstance(val, list), " ".join(auxp.split())
                self._data[name] = val
            else:
                auxp = "You are trying to set the \
                parameter {0} that have to be a single number or other single type \
                of data with a list in the class {1}".format(name, \
                self.__class__.__name__)
                assert not isinstance(val, list), " ".join(auxp.split())
                self._data[name] = val
        else:
            auxp = "No valid key {0} has been passed to the function \
            set_element in the class {1}. The valid keys for this class are as \
            follows \n {2}".format(name, self.__class__.__name__, \
            self._data.keys()) 
            assert " ".join(auxp.split())

class Branch(_CommonMethods):
    ''' Electricity branch - This could be information for either transmission \
        lines or two winding transformers '''
    def __init__(self):
        # Data that have to be single numbers or names
        aux = ['shunt_susceptance', 'resistance', 'status', 'reactance',
               'number', 'long_term_thermal_limit', 
               'short_term_thermal_limit', 'emergency_thermal_limit',
               'non_technical_losses_fix', 'position',
               'contingency_n-1']
        self._data = {}
        for x in aux:
            self._data[x] = None

        # Data that have to be a list of numbers or names
        aux =  ['bus_number', 'bus_position']
        __data2 = {}
        for x in aux:
            __data2[x] = []
        
        self._data.update(__data2)
        del __data2

class Transformers(Branch):
    ''' Two winding transformer class '''
    def __init__(self):
        super().__init__()
        # Basic _data
        aux = ['max_phase_shift_angle', 'min_phase_shift_angle', 'tap']
        __data2 = {}
        for x in aux:
            __data2[x] = None
        
        self._data.update(__data2)
        del __data2

class TransmissionLine(Branch):
    ''' Transmission Line class '''
    def __init__(self):
        super().__init__()

class Bus(_CommonMethods):
    ''' Electricity bus '''
    def __init__(self):
        # Data that have to be single numbers or names
        aux = ['voltage_kv', 'shunt_susceptance', 'area', 'type', 'position_x', 
               'position_y', 'shunt_conductance', 
               'active_power_demand_peak', 'reactive_power_demand_peak', 
               'position', 'name', 'number','initial_voltage_magnitude', 
               'initial_voltage_angle', 'maximum_voltage_magnitude', 
               'minimum_voltage_magnitude', 'zone', 'non_technical_losses_fix',
               'load_type']
        self._data = {}
        for x in aux:
            self._data[x] = None

        # Data that have to be a list of numbers or names
        aux =  ['contingency_n-1', 
                'transmissionline_position', 'transmissionline_number',
                'conv_position', 'conv_number',
                'hydro_position', 'hydro_number',
                'RES_position', 'RES_number',
                'transformers_position', 'transformers_number']
        __data2 = {}
        for x in aux:
            __data2[x] = []
        
        self._data.update(__data2)
        del __data2

class ElectricityNetwork(_CommonMethods):
    ''' Electricity network '''
    def __init__(self, **kwargs):
        ''' General electricity network settings 
        In case the user wants to initialise some or all the objects then \
        the some or all the following arguments have to be passed in kwargs:
        'nobuses' -> Number of buses
        'noconv' -> Number of conventional generators
        'nohydro' -> Number of hydro generators
        'nores' -> Number of RES generators
        'notlines' -> Number of transmission lines
        'notrafos' -> Number of transformers
        '''
        self._data = {
            'baseMVA': None,
            'Slack': None,
            'Security': [],  # list of N-1 cases to consider
            'Threewindinginitial' : None, # initial bus for three winding trafos
            'voltagethreewindingtrafos' : 1000000 # voltage for artificial nodes 
                                                  # in three winding trafos
            }
        
        self.__objects = {
            'bus' : [], # list of Bus objects
            'conv' : [], # list of conventional generator objects
            'hydro' : [], # list of hydro generator objects
            'RES' : [], # list of RES generator objects
            'transformers' : [], # list of two winding trafo objects
            'transmissionline' : [] # list of transmission line objects
            }
        
        # This list should contain the names of all series elements
        self.__series_elements = [
            'transmissionline', 'transformers'
        ]

        # This list should contain the names of all generator types
        self.__generation_types = [
            'conv', 'RES', 'hydro'
        ]

        self.__parameters_elements = [
            'baseMVA', 'Slack', 'Security', 'Threewindinginitial',
            'voltagethreewindingtrafos'
        ]

        # Initialise bus object
        if 'nobuses' in kwargs.keys():
            self.__objects['bus'] = [Bus() for _ in
                range(kwargs.pop('nobuses'))]
        else:
            self.__objects['bus'] = [Bus()]

        # Initialise Conventional generator object
        if 'noconv' in kwargs.keys():
            self.__objects['conv'] = [Conventional() for _ in
                range(kwargs.pop('noconv'))]
        else:
            self.__objects['conv'] = [Conventional()]

        # Initialise hydro-electrical generator object
        if 'nohydro' in kwargs.keys():
            self.__objects['hydro'] = [Hydropower() for _ in
                range(kwargs.pop('nohydro'))]
        else:
            self.__objects['hydro'] = [Hydropower()]

        # Initialise RES generator object
        if 'nores' in kwargs.keys():
            self.__objects['RES'] = [RES() for _ in
                range(kwargs.pop('nores'))]
        else:
            self.__objects['RES'] = [RES()]
                
        # Initialise transmission line object
        if 'notlines' in kwargs.keys():
            self.__objects['transmissionline'] = [TransmissionLine() for _ in
                    range(kwargs.pop('notlines'))]
        else:
            self.__objects['transmissionline'] = [TransmissionLine()]

        # Initialise transformer object
        if 'notrafos' in kwargs.keys():
            self.__objects['transformers'] = [Transformers() for _ in
                range(kwargs.pop('notrafos'))]
        else:
            self.__objects['transformers'] = [Transformers()]
    
    def copy_electricity_network_data(self):
        ''' This returns a copy of the electricity network object'''
        return self.__dict__

    def delete_objects(self, name=None, pos=None):
        ''' This function delete an object or a list of objects for the 
        requested object "name". 
        - The function delete either a list with the positions indicated 
        in "pos" or the object of the position "pos"
        - "pos" needs to indicate a position between zero and the size of the 
        '''
        auxp = "No valid name has been passed to the function delete_objects \
            in the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        auxp = "No valid position has been passed to the function \
            delete_objects in the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self.__objects:
            copy_class = copy.deepcopy(self.__objects[name][0])
            copy_class.__init__()
            if isinstance(pos, list):
                self.__objects[name] = [i for j, i in \
                    enumerate(self.__objects[name]) if j not in pos]
            else:
                self.__objects[name].pop(pos)
            if self.__objects[name] == []:
                self.__objects[name] = [copy_class]
            self.__update_positions_objects(name)
            if name == 'bus':
                 # Updating the positions related to the bus object
                for xobj in self.__objects.keys():
                    if xobj != 'bus':
                        self.__update_relative_position_objects(\
                            name_object1='bus', \
                            name_element1='number', name_object2=xobj, \
                            name_element2='bus_number', 
                            name_position_element='bus_position')
            else:
                # Updating the positions of other objects in the bus object
                self.__update_relative_position_objects(name_object1=name, \
                    name_element1='number', name_object2='bus', \
                    name_element2=name+'_number', \
                    name_position_element=name+'_position')
        else:
            auxp = "No valid key {0} has been passed to the function \
            delete_objects in the class {1}. The valid keys for this class are \
            as follows: \n {2}".format(name, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())

    def get_generation_types_names(self):
        ''' This function returns a list of all generation types considered in 
        the electricity network class'''
        return self.__generation_types

    def get_no_objects(self, name=None):
        ''' This function returns the number of objects in a list '''
        auxp = "No valid name has been passed to the function get_no_objects \
            in the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self.__objects:
            if self.__objects[name][0].get_element(name='position') != None:
                return len(self.__objects[name])
            else:
                return 0
        else:
            auxp = "No valid key {0} has been passed to the function \
            get_no_objects in the class {1}. The valid keys for this class are \
            as follows: \n {2}".format(name, self.__class__.__name__, \
            self._data.keys()) 
            assert " ".join(auxp.split())

    def get_objects(self, name=None, pos=':'):
        ''' This function returns an object or a list of objects for the 
        requested object "name". 
        - The function returns either a list with the positions indicated 
        in "pos" or the object of the position "pos"
        - "pos" needs to indicate a position between zero and the size of the 
        list
        - The function returns by default the whole list of objects'''
        auxp = "No valid name has been passed to the function get_objects in \
            the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        if name in self.__objects:
            if pos == ':':
                return self.__objects[name]
            elif isinstance(pos, list):
                aux = []
                for aux1 in pos:
                    aux.append(self.__objects[name][aux1])
                return aux
            else:
                return self.__objects[name][pos]
        else:
            auxp = "No valid key {0} has been passed to the function \
            get_objects in the class {1}. The valid keys for this class are as \
            follows: \n {2}".format(name, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())
    
    def get_object_elements(self, name_object=None, name_element=None, \
        pos_element=':', pos_object=None):
        ''' This function returns the value or a list of values for the 
        requested parameter "name_element" of the object "name_obj" in position
        "pos_object". 
        - If the parameter is a list then the function returns either a list 
        with the positions indicated in "pos_element" or the value of the 
        position "pos_element"
        - "pos_element" needs to indicate a position between zero and the size 
        of the list of parameters
        - "pos_object" needs to indicate a position between zero and the size 
        of the list of objects
        - If the parameters is a value then "pos" can be ignored
        - The function returns by default the whole list of parameters '''

        auxp = "No valid object name has been passed to the function \
            get_objects_elements in the class \
            {0}".format(self.__class__.__name__)
        assert name_object is not None, " ".join(auxp.split())
        auxp = "No valid object position has been passed to the function \
            get_objects_elements in \
            the class {0}".format(self.__class__.__name__)
        assert pos_object is not None, " ".join(auxp.split())
        if name_object in self.__objects:
            return self.__objects[name_object][pos_object].get_element(\
                name=name_element, pos=pos_element)
        else:
            auxp = "No valid key {0} has been passed to the function \
            get_object_elements in the class {1}. The valid keys for this class are as \
            follows: \n {2}".format(name_object, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())
    
    def get_parameters_list(self):
        ''' This function returns a list parameters considered in 
        the electricity network class'''
        return self.__parameters_elements

    def get_series_elements_names(self):
        ''' This function returns a list of all series elements considered in 
        the electricity network class'''
        return self.__series_elements

    def set_objects(self, name=None, list_obj=None):
        ''' This function set a list of a specific object indicated in "name". 
        - This function rewrite the list of a specific object '''
        auxp = "No valid name has been passed to the function set_element in \
            the class {0}".format(self.__class__.__name__)
        assert name is not None, " ".join(auxp.split())
        auxp = "No valid list of objects has been passed to the function \
            set_element in \
            the class {0}".format(self.__class__.__name__)
        assert list_obj is not None, " ".join(auxp.split())
        if name in self.__objects:
            if isinstance(self.__objects[name], list):
                auxp = "You are trying to set the list of\
                    objects {0} with different type of \
                    data element in the class {1}".format(name, \
                    self.__class__.__name__)
                assert isinstance(list_obj, list), " ".join(auxp.split())
                auxp = "The objects inside the list are not instances of the \
                    the class {0}".format(\
                    self.__objects[name][0].__class__.__name__)
                assert isinstance(list_obj[0], \
                    self.__objects[name][0].__class__), \
                    " ".join(auxp.split())
                self.__objects[name] = list_obj
        else:
            auxp = "No valid key {0} has been passed to the function \
            set_objects in the class {1}. The valid keys for this class are as \
            follows \n {2}".format(name, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())

    def set_object_elements(self, name_object=None, name_element=None, \
        val=None, pos_object=None):
        ''' This function set the value or a list of values for the 
        requested parameter "name_element" of the object "name_obj" in position
        "pos_object". 
        - This function rewrite the existing values or value stored in the 
        parameter
        - "pos_object" needs to indicate a position between zero and the size 
        of the list of objects '''

        auxp = "No valid object name has been passed to the function \
            get_objects_elements in \
            the class {0}".format(self.__class__.__name__)
        assert name_object is not None, " ".join(auxp.split())
        auxp = "No valid object position has been passed to the function \
            get_objects_elements in \
            the class {0}".format(self.__class__.__name__)
        assert pos_object is not None, " ".join(auxp.split())
        if name_object in self.__objects:
            self.__objects[name_object][pos_object].set_element(\
                name=name_element, val=val)
        else:
            auxp = "No valid key {0} has been passed to the function \
            set_object_elements in the class {1}. The valid keys for this class are as \
            follows: \n {2}".format(name_object, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())

    def set_electricity_network_data(self, ob=None):
        ''' set the data of the electricity network object '''
        assert isinstance(ob, ElectricityNetwork), "Incorrect object \
            passed to set the Electricity Network data"
        for xkey in self.__objects.keys():
            self.set_objects(name=xkey, list_obj=ob.get_objects(name=xkey))
        for xkey in self.get_parameters_list():
            self.set_element(name=xkey,val=ob.get_element(name=xkey))
    
    def update_all_positions(self):
        ''' Update the position of all nodes, transmission lines, etc. '''
        all_obj = ['bus']
        all_obj.extend(self.get_generation_types_names())
        all_obj.extend(self.get_series_elements_names())
        for xobj in all_obj:
            self.__update_positions_objects(name_object=xobj)
        # Updating the positions related to the bus object
        for xobj in all_obj:
            if xobj != 'bus':
                self.__update_relative_position_objects(name_object1='bus', \
                    name_element1='number', name_object2=xobj, \
                    name_element2='bus_number', 
                    name_position_element='bus_position')
        # Updating the positions of other objects in the bus object
        for xobj in self.get_series_elements_names():
            self.__update_relative_position_objects(name_object1=xobj, \
                name_element1='number', name_object2='bus', \
                name_element2=xobj+'_number', \
                name_position_element=xobj+'_position')
        for xobj in self.get_generation_types_names():
            self.__update_relative_position_objects(name_object1=xobj, \
                name_element1='bus_number', name_object2='bus', \
                name_element2=xobj+'_number', \
                name_position_element=xobj+'_position')

    def __update_relative_position_objects(self, name_object1=None, \
        name_element1=None, name_object2=None, name_element2=None, 
        name_position_element=None):
        ''' This function update the relative position of the object 
        "name_object1" in the object "name_object2". 
        - "name_element1" and "name_element2" are elements in the objects that 
        are compared to update the positions 
        - "name_element1" is always a single parameter
        - "name_element2" can be a list or a single value
        - "name_position_element" is the name of the elements in the object 
        "name_object2" that will be updated '''
        auxp = "No valid object name 'name_object1' has been passed to the \
            function __update_relative_position_objects in \
            the class {0}".format(self.__class__.__name__)
        assert name_object1 is not None, " ".join(auxp.split())
        auxp = "No valid object name 'name_object2' has been passed to the \
            function __update_relative_position_objects in \
            the class {0}".format(self.__class__.__name__)
        assert name_object2 is not None, " ".join(auxp.split())
        aux = self.__objects[name_object2][0].get_element(name='position')
        aux1 = self.__objects[name_object1][0].get_element(name='position')
        # If any of the objects does not have values then there is nothing to
        # update
        if aux == None or aux1 == None:
            if name_object2 not in self.get_generation_types_names():
                for xobj2 in self.__objects[name_object2]:
                    xobj2.set_element(name=name_position_element, val=[])
            return

        if name_object1 in self.__objects and name_object2 in self.__objects:
            for xobj2 in self.__objects[name_object2]:
                aux1 = xobj2.get_element(name=name_element2)
                if isinstance(aux1, list):
                    aux_bool = [False for _ in aux1]
                    aux3 = []
                    for xobj1 in self.__objects[name_object1]:
                        for aux2 in range(len(aux1)):
                            if aux1[aux2] == xobj1.get_element(\
                                name=name_element1) and not aux_bool[aux2]:
                                aux_bool[aux2] = True
                                aux3.append(xobj1.get_element(name='position'))
                                break
                    xobj2.set_element(name=name_position_element, val=aux3)
                else:
                    for xobj1 in self.__objects[name_object1]:
                        if aux1 == xobj1.get_element(name=name_element1):
                            xobj2.set_element(name=name_position_element, \
                                val=xobj1.get_element(name='position'))
                            break
        else:
            auxp = "No valid key {0} has been passed to the function \
            set_objects in the class {1}. The valid keys for this class are as \
            follows \n {2}".format(name_object1, self.__class__.__name__, \
            self.__objects.keys()) 
            assert name_object1 in self.__objects, " ".join(auxp.split())

            auxp = "No valid key {0} has been passed to the function \
            set_objects in the class {1}. The valid keys for this class are as \
            follows \n {2}".format(name_object2, self.__class__.__name__, \
            self.__objects.keys()) 
            assert name_object2 in self.__objects, " ".join(auxp.split())
    
    def __update_positions_objects(self, name_object=None):
        ''' This function update the position of the list of the object 
        "name_obj". 
        - This function rewrite the existing values of the positions '''

        auxp = "No valid object name has been passed to the function \
            get_objects_elements in \
            the class {0}".format(self.__class__.__name__)
        assert name_object is not None, " ".join(auxp.split())
        if name_object in self.__objects:
            if self.__objects[name_object][0].get_element(name='position') != \
                None:
                aux=0
                for xn in self.__objects[name_object]:
                    xn.set_element(name='position', val=aux)
                    aux += 1
        else:
            auxp = "No valid key {0} has been passed to the function \
            set_object_elements in the class {1}. The valid keys for this class are as \
            follows: \n {2}".format(name_object, self.__class__.__name__, \
            self.__objects.keys()) 
            assert " ".join(auxp.split())

class GenClass(_CommonMethods):
    ''' Core generation class '''
    def __init__(self):
        # Basic _data
        aux = [ 'maximum_active_power_generation', 
                'minimum_active_power_generation', 
                'maximum_reactive_power_generation',
                'minimum_reactive_power_generation', 'ramp', 'baseload',
                'model', 'shutdown_cost', 'startup_cost', 'position',
                'uncertainty', 'bus_number', 'bus_position', 'status']
        self._data = {}
        for x in aux:
            self._data[x] = None
        
        aux =  ['cost_function_parameters', 
        'piecewise_linearization_parameters']
        __data2 = {}
        for x in aux:
            __data2[x] = []
        
        self._data.update(__data2)
        del __data2

class Conventional(GenClass):
    ''' Conventional generator '''
    def __init__(self):
        self.cooling = {}
        self.cooling['Flag'] = False
        self.cooling['Gen2Der'] = []
        self.cooling['Derate'] = None
        super().__init__()

class Hydropower(GenClass):
    ''' Hydropower generator '''
    def __init__(self):
        super().__init__()

class RES(GenClass):
    ''' RES generation '''
    def __init__(self):
        super().__init__()



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 25 11:23:14 2018

Pyene for Printing outputs

@author: Dr Eduardo Alejandro Martínez Ceseña
https://www.researchgate.net/profile/Eduardo_Alejandro_Martinez_Cesena
"""
import numpy as np

class PrintClass:
    def __init__(self, m, EN):
        ''' Initialize for pyomo, pypsa or glpk class '''
        
        self.data = {
                'GenerationP': None,
                'GenerationQ': None,
                'GenerationUC': None,
                'GenerationCost': None,
                'Line_FlowP0': None,
                'Line_FlowP1': None,
                'Line_FlowQ0': None,
                'Line_FlowQ1': None,
                'Line_LossP': None,
                'Line_LossQ': None,
                'Voltage_pu': None,
                'Voltage_ang': None,
                'Pumps': None,
                'Curtailment': None,
                'Services': None,
                'OF': None
                }
        
        from pyomo.core import ConcreteModel
        import pypsa

        if isinstance(m, ConcreteModel):
            self.type = 1
        elif isinstance(m[0], pypsa.Network):
            self.type = 2
        else:
            print('Model not recognised')

        # Initialize network variables
        from .main import pyeneClass
        from .pyeneN import ENetworkClass
        if isinstance(EN, pyeneClass):
            OFaux = EN._Calculate_OFaux()
            self.initializeNM(m, EN.NM, OFaux)
        elif isinstance(EN, ENetworkClass):
            OFaux = np.ones(len(EN.connections['set']),dtype=float)
            self.initializeNM(m, EN, OFaux)

    def initializeNM(self, m, NM, OFaux):
        Noh = len(NM.connections['set'])
        self.data['GenerationP'] = \
            np.zeros((Noh*NM.Gen.get_NoGen(), NM.settings['NoTime']),
                     dtype=float)
        self.data['GenerationQ'] = \
            np.zeros((Noh*NM.Gen.get_NoGen(), NM.settings['NoTime']),
                     dtype=float)
        self.data['GenerationCost'] = \
            np.zeros((Noh*NM.Gen.get_NoGen(), NM.settings['NoTime']),
                     dtype=float)
        self.data['GenerationUC'] = \
            np.ones((Noh*NM.Gen.get_NoGen(), NM.settings['NoTime']),
                     dtype=bool)
        self.data['Voltage_pu'] = \
            np.ones((Noh*NM.NoBuses, NM.settings['NoTime']),
                     dtype=float)
        self.data['Voltage_ang'] = \
            np.ones((Noh*NM.NoBuses, NM.settings['NoTime']),
                     dtype=float)
        self.data['Pumps'] = np.zeros((Noh*(NM.pumps['Number']+1),
                                      NM.settings['NoTime']), dtype=float)
        self.data['Curtailment'] = \
            np.zeros((Noh*NM.NoFea, NM.settings['NoTime']), dtype=float)
        self.data['Services'] = np.zeros((Noh*NM.p['GServices'],
                                        NM.settings['NoTime']), dtype=float)
        self.data['Line_FlowP0'] = \
            np.zeros((Noh*NM.NoBranch, NM.settings['NoTime']), dtype=float)
        self.data['Line_FlowP1'] = \
            np.zeros((Noh*NM.NoBranch, NM.settings['NoTime']), dtype=float)
        self.data['Line_FlowQ0'] = \
            np.zeros((Noh*NM.NoBranch, NM.settings['NoTime']), dtype=float)
        self.data['Line_FlowQ1'] = \
            np.zeros((Noh*NM.NoBranch, NM.settings['NoTime']), dtype=float)
        self.data['Line_LossP'] = \
            np.zeros((Noh*NM.connections['Branches'], NM.settings['NoTime']),
                     dtype=float)
        self.data['Line_LossQ'] = \
            np.zeros((Noh*NM.connections['Branches'], NM.settings['NoTime']),
                     dtype=float)

        if self.type == 1:
            # Pyomo - DC model
            for x1 in range(Noh*NM.Gen.get_NoGen()):
                for x2 in range(NM.settings['NoTime']):
                     self.data['GenerationP'][x1][x2] = \
                            m.vNGen[x1, x2].value*NM.ENetwork.get_Base()

            # UC
            if 'vNGen_Bin' in m.__dir__():
                for x1 in range(Noh*NM.Gen.get_NoGen()):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['GenerationUC'][x1][x2] = \
                            m.vNGen_Bin[x1, x2].value

            # Voltage angle
            if 'vNVolt' in m.__dir__():
                for x1 in range(Noh*NM.NoBuses):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Voltage_ang'][x1][x2] = \
                                m.vNVolt[x1, x2].value

            # Pumps
            if 'vNPump' in m.__dir__():
                for x1 in range(Noh*(NM.pumps['Number']+1)):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Pumps'][x1][x2] = \
                                m.vNPump[x1, x2].value*NM.ENetwork.get_Base()

            # Dummy genertors / curtailment / feasibility constraints
            if 'vNFea' in m.__dir__():
                for x1 in range(Noh*NM.NoFea):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Curtailment'][x1][x2] = \
                                m.vNFea[x1, x2].value*NM.ENetwork.get_Base()

            # Services
            if 'vNServ' in m.__dir__():
                for x1 in range(Noh*NM.p['GServices']):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Services'][x1][x2] = \
                                m.vNServ[x1, x2].value*NM.ENetwork.get_Base()

            # Lines
            if 'vNFlow' in m.__dir__():
                for x1 in range(Noh*NM.NoBranch):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Line_FlowP0'][x1][x2] = \
                            m.vNFlow[x1, x2].value*NM.ENetwork.get_Base()

            # Losses
            if 'vNLoss' in m.__dir__():
                for x1 in range(Noh*NM.connections['Branches']):
                    for x2 in range(NM.settings['NoTime']):
                        self.data['Line_LossP'][x1][x2] = \
                        m.vNLoss[x1, x2].value*NM.ENetwork.get_Base()

            # OF
            self.data['OF'] = m.OF.expr()

        elif self.type ==2:
            # pypsa - AC model
            for xh in range(Noh):
                # Generators
                for xg in range(NM.Gen.get_NoGen()):
                    x1 = NM.connections['Generation'][xh]+xg
                    self.data['GenerationP'][x1][:] = \
                        getattr(m[xh].generators_t.p, 'Gen{}'.format(xg+1))
                    self.data['GenerationQ'][x1][:] = \
                        getattr(m[xh].generators_t.q, 'Gen{}'.format(xg+1))

                # Generation costs
                NM.s['Gen'] = range(NM.Gen.get_NoGen()) # xg
                range(NM.Gen.get_NoPieces()) # xc
                NM.s['Tim'] = range(NM.settings['NoTime']) # xt
                NM.s['Con'] = NM.connections['set'] # xh
                ConC = NM.connections['Cost'][xh]
                ConG = NM.connections['Generation'][xh]
                for xg in range(NM.Gen.get_NoGen()):
                    for xc in range(NM.Gen.get_NoPieces()):
                        (flg, x1, x2, M1, M2) = \
                        NM.Gen.cNEGenC_Auxrule(xg, xc, ConC, ConG)
                        if flg:
                            for xt in range(NM.settings['NoTime']):
                                w = NM.scenarios['Weights'][xt]
                                self.data['GenerationCost'][x1][xt] = \
                                    max(self.data['GenerationCost'][x1][xt],
                                        self.data['GenerationQ'][x1][xt] *
                                        M1*w+M2)
                # Buses
                for xn in range(NM.ENetwork.get_NoBus()):
                    x1 = NM.connections['Voltage'][xh]+xn
                    self.data['Voltage_pu'][x1][:] = \
                        getattr(m[xh].buses_t.v_mag_pu, 'Bus{}'.format(xn+1))
                    self.data['Voltage_ang'][x1][:] = \
                        getattr(m[xh].buses_t.v_ang, 'Bus{}'.format(xn+1))

                # Lines
                xb = 0
                for ob in NM.ENetwork.Branch:
                    # Is this a line?
                    if ob.get_Tap() == 0:
                        aux = 'Line{}'.format(ob.get_Number())
                        P0 = getattr(m[xh].lines_t.p0, aux)
                        P1 = getattr(m[xh].lines_t.p1, aux)
                        Q0 = getattr(m[xh].lines_t.q0, aux)
                        Q1 = getattr(m[xh].lines_t.q1, aux)
                    else:  # Transformer
                        aux = 'Trs{}'.format(ob.get_Number())
                        P0 = getattr(m[xh].transformers_t.p0, aux)
                        P1 = getattr(m[xh].transformers_t.p1, aux)
                        Q0 = getattr(m[xh].transformers_t.q0, aux)
                        Q1 = getattr(m[xh].transformers_t.q1, aux)

                    x1 = NM.connections['Flow'][xh]+xb
                    self.data['Line_FlowP0'][x1][:] = P0
                    self.data['Line_FlowP1'][x1][:] = P1
                    self.data['Line_FlowQ0'][x1][:] = Q0
                    self.data['Line_FlowQ1'][x1][:] = Q1
                    self.data['Line_LossP'][x1][:] = P0+P1
                    self.data['Line_LossQ'][x1][:] = Q0+Q1
                    xb += 1

            # OF
            self.data['OF'] = 0
            for xh in range(Noh):
                for xg in range(NM.Gen.get_NoGen()):
                    x1 = NM.connections['Generation'][xh]+xg
                    self.data['OF'] += self.data['GenerationCost'][x1][xt] * \
                        OFaux[xh]

    def get_Curtailment(self, x1, x2):
        return self.data['Curtailment'][x1][x2]

    def get_GenerationP(self, x1, x2):
        return self.data['GenerationP'][x1][x2]

    def get_GenerationQ(self, x1, x2):
        return self.data['GenerationQ'][x1][x2]

    def get_GenerationUC(self, x1, x2):
        return self.data['GenerationUC'][x1][x2]

    def get_Line_FlowP0(self, x1, x2):
        return self.data['Line_FlowP0'][x1][x2]

    def get_Line_FlowP1(self, x1, x2):
        return self.data['Line_FlowP1'][x1][x2]

    def get_Line_FlowQ0(self, x1, x2):
        return self.data['Line_FlowQ0'][x1][x2]

    def get_Line_FlowQ1(self, x1, x2):
        return self.data['Line_FlowQ1'][x1][x2]

    def get_Line_LossP(self, x1, x2):
        return self.data['Line_LossP'][x1][x2]

    def get_Line_LossQ(self, x1, x2):
        return self.data['Line_LossQ'][x1][x2]

    def get_OF(self):
        return self.data['OF']

    def get_Pumps(self, x1, x2):
        return self.data['Pumps'][x1][x2]

    def get_Services(self, x1, x2):
        return self.data['Services'][x1][x2]

    def get_Voltage_pu(self, x1, x2):
        return self.data['Voltage_pu'][x1][x2]

    def get_Voltage_ang(self, x1, x2):
        return self.data['Voltage_ang'][x1][x2]

